#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <avr/pgmspace.h>

#define TXD _BV(PB1)
#define BAUD_RATE 9600

volatile unsigned char txState;
volatile unsigned char txByte;

volatile unsigned char conversionReady;
volatile unsigned long conversionCount;

ISR(TIM0_COMPA_vect) {
	switch(txState) {
		case 0:
		case 1:
			// idle or stop bit
			PORTB |= TXD;
			break;
		case 10:
			// start bit
			PORTB &= (unsigned char)~TXD;
			break;
		default:
			// data bit
			if((txByte >> (9 - txState)) & 0x1) PORTB |= TXD;
			else PORTB &= (unsigned char)~TXD;
			break;
	}
	if(txState) txState--;
}


ISR(ADC_vect) {
	conversionReady = 1;
	conversionCount++;
	ADCSRA &= (unsigned char)~_BV(ADEN);
}

void sendChar(unsigned char c) {
	while(txState) {
		set_sleep_mode(SLEEP_MODE_IDLE);
		sleep_cpu();	// wait for TX to finish
	}
	txByte = c;
	txState = 10;
}

unsigned long mags[10] PROGMEM = {
	1000000000,
	100000000,
	10000000,
	1000000,
	100000,
	10000,
	1000,
	100,
	10,
	1
};

void sendLong(unsigned long num) {
	unsigned char i, n, first = 1;
	unsigned long m;
	for(i = 0; i < 10; i++) {
		m = pgm_read_dword(&mags[(unsigned)i]);
		for(n = 0; num >= m; num -= m) n++;
		if(first && n < 1 && i < 9) continue;
		first = 0;
		sendChar('0' + n);
	}
}

void sendHex(unsigned char c) {
	if(((c >> 4) & 0xf) < 10) sendChar('0' + ((c >> 4) & 0xf));
	else sendChar('A' + ((c >> 4) & 0xf) - 10);
	if((c & 0xf) < 10) sendChar('0' + (c & 0xf));
	else sendChar('A' + (c & 0xf) - 10);
}

void init() {
	txState = 0;	// idle

	TCCR0A = _BV(WGM01);		// Clear timer on compare match mode
	TCCR0B = _BV(CS02);		// Clock prescale by 256
	OCR0A = F_CPU/256/BAUD_RATE;	// Compare match at baud-rate
	TIMSK0 = _BV(OCIE0A);		// Enable compare match interrupts
	
	DDRB |= TXD;	// TX pin output enable
	PORTB |= TXD;	// idle

	conversionCount = 0;

	ADMUX = _BV(MUX0);		// ADC1/PB2 as input, LSB aligned, Vcc ref
	ADCSRA = _BV(ADEN) | _BV(ADIE) | _BV(ADPS2) | _BV(ADPS1) | _BV(ADPS0);	// enable ADC with interrupt notification and /128 prescale

	sleep_enable();

	sei();
}

int getADC(unsigned char channel) {
	int adcVal;

	switch(channel) {
		case 0:
			ADMUX = _BV(MUX0);
			break;
		case 1:
			ADMUX = _BV(MUX1);
			break;
		case 2:
			ADMUX = _BV(MUX1) | _BV(MUX0);
			break;
	}
	
	conversionReady = 0;
	while(!conversionReady) {
		ADCSRA |= _BV(ADEN);
		set_sleep_mode(SLEEP_MODE_ADC);
		sleep_cpu();
	}

	adcVal = ADCL;
	adcVal |= ADCH << 8;

	return adcVal;
}

int main() {
	int adc0, adc1, adc2;

	init();

	while(1) {
		while(txState) {
			set_sleep_mode(SLEEP_MODE_IDLE);
			sleep_cpu();
		}


		adc0 = getADC(0);
		adc1 = getADC(1);
		adc2 = getADC(2);

		sendLong(conversionCount);
		sendChar(' ');
		
		sendLong(adc0);
		sendChar(' ');
		sendLong(adc1);
		sendChar(' ');
		sendLong(adc2);

		sendChar('\r');
		sendChar('\n');
	}

	return 0;  // unreached
}
